using System;
using System.Collections.Generic;
using System.Text;

namespace Sketcher
{
    abstract class InputBehavior : Behavior
    {
        public InputBehavior(Object target) : base(target) 
        {
            Connect();
        }


        //! connect this behavior to the world
        override public void Connect() {
            InputBehaviorManager.Get.Register( this );
        }

        //! disconnect this behavior from the world
        override public void Disconnect( ) {
            InputBehaviorManager.Get.DeRegister( this );
        }

        //! called when input device(s) move w/o being activated (note that some devices don't actually provide this)
        virtual public void OnHover(Vector2f v) {
        }


        public enum InputCaptureType
        {
            NeverCapture,                   //! will never want to capture input
            FirstVertexOrNeverCapture,      //! if doesn't capture on first stroke vertex, will never capture
            MayCapture                      //! might capture input
        };

        virtual public InputCaptureType CaptureType {
            get { return InputCaptureType.NeverCapture; }
        }

        //! return value in [0,1] to indicate desire for capturing stroke
        virtual public float GetCaptureScore(Stroke s) {
            return 0.0f;
        }
        //! return this value from GetCaptureScore to prevent any behaviors from capturing input
        public static float PREVENT_CAPTURE = float.MaxValue;

        //! return true to continue capture (although if you return false, why did you say you wanted to capture?)
        virtual public bool OnCaptureStrokeStart(Stroke s) {
            return true;
        }

        //! return true to continue capture, false to stop capture
        virtual public bool OnCapturedStrokeUpdate(Stroke s) {
            return true;
        }

        //! return true if stroke was consumed, or false to allow processing as gesture
        virtual public bool OnCaptureStrokeEnd(Stroke s) {
            return true;
        }


        //! return cross "factor"  (length of cross inside region)
        //virtual public float GetCrossScore(LineStroke s) {
        //    return 0.0f;
        //}

        //! return cross features (scores for different measures of "crossed"-ness)
        //! return null or empty FeatureSet to ignore event
        virtual public FeatureVector GetCrossFeatures(LineStroke s) {
            return null;     
        }

        //! handle cross event
        virtual public void OnCross(LineStroke s) {

        }


        //! return circle "factor" 
        //! (should be percentage of circle inside region, but currently just 1/0 depending on containment of center)
        virtual public float GetCircleScore(CircleStroke s) {
            return 0.0f;
        }

        //! handle circle event
        virtual public void OnCircle(CircleStroke s) {

        }



        //! return click "score"
        virtual public float GetClickScore(ClickStroke s) {
            return 0.0f;
        }

        //! handle click event
        virtual public void OnClick(ClickStroke s) {
        }


        //! return general stroke "score"
        virtual public float GetGeneralStrokeScore( Stroke s ) {
            return 0.0f;
        }

        //! handle general stroke event
        virtual public void OnGeneralStroke( Stroke s ) {
        }

    }


    /*
     * "versions" of class with different types of target
     */

    abstract class ElementInputBehavior : InputBehavior 
    {
        public ElementInputBehavior(Element target)
            : base(target) {
        }

        public Element TargetElement {
            get { return Target as Element; }
        }

        //! behaviors are only considered for enabled items
        override public bool Enabled {
            get { return Element.VisibleForRendering(TargetElement.Visibility); }
        }
    }


    abstract class ItemInputBehavior : ElementInputBehavior
    {
        public ItemInputBehavior(Item target)
            : base(target) {
        }

        public Item TargetItem {
            get { return Target as Item; }
        }
    }



    abstract class WidgetInputBehavior : ElementInputBehavior
    {
        public WidgetInputBehavior(Widget target)
            : base(target) {
        }

        public Widget TargetWidget {
            get { return Target as Widget; }
        }

    }
}
