using System;
using System.Collections.Generic;
using System.Text;

namespace Sketcher {

    /*
     * Base class for UI Elements ( Item and Widget )
     */

    abstract class Element : BehaviorTarget, Renderable
    {
        // properties for Renderable
        protected ColorRGBA m_fgColor;
        protected ColorRGBA m_bgColor;
        protected float m_fOpacity;
        protected VisibilityState m_Visibility;

        // position information
        protected Transform m_transform;
        protected Size2f m_size;

        protected Element m_parent;

        public Element() {
            m_transform = new Transform();
            m_size = new Size2f(false);
            m_parent = null;
        }

        ~Element( ) {
            System.Diagnostics.Debug.WriteLine("Element destructor!");
        }


        //! disconnect element from the world
        virtual public void Disconnect( ) {
            Parent = null;
            Behaviors.Disconnect();
            WidgetManager.Get.DisconnectWidgets( this );
        }

        //! reconnect element to world. Note that this does not reconnect Parent or any Widgets
        virtual public void Reconnect() {
            Behaviors.Connect();
            // [TODO] restore widgets (?)
        }


        virtual public Transform Transformation {
            get { return m_transform; }
        }
        virtual public Size2f Size {
            get { return m_size; }
        }
        virtual public Box2f Bounds {
            get { return Transformation.Apply(Size.ToBox()); }
        }

        virtual public Element Parent {
            get { return m_parent; }
            set { m_parent = value; }
        }

        //! fire transition animation to show this element
        virtual public void Show() {
            TransitionBehaviorManager.Get.Show(this);
        }

        //! fire transition animation to hide this element
        virtual public void Hide() {
            TransitionBehaviorManager.Get.Hide(this);
        }


        virtual public ColorRGBA AlphaForegroundColor {
            get {  ColorRGBA fg = ForegroundColor;
                   fg.A *= Opacity;
                   return fg; 
            }
        }
        virtual public ColorRGBA AlphaBackgroundColor {
            get {
                ColorRGBA bg = BackgroundColor;
                bg.A *= Opacity;
                return bg;
            }
        }



#region Renderable Members

        virtual public VisibilityState Visibility {
            get {  return m_Visibility;  }
            set {  m_Visibility = value; NotifyElementVisibilityChanged(); }
        }

        virtual public float Opacity {
            get { return (Parent == null ) ? m_fOpacity : m_fOpacity * Parent.Opacity; }
            set { m_fOpacity = value; NotifyElementChanged(); }
        }

        virtual public ColorRGBA ForegroundColor {
            get { return m_fgColor; }
            set { m_fgColor = value; NotifyElementChanged(); }
        }
        virtual public ColorRGBA BackgroundColor {
            get { return m_bgColor; }
            set { m_bgColor = value; NotifyElementChanged(); }
        }


        /*
         *  element changed 
         */
        public delegate void ElementChangedHandler(Element e);
        public event ElementChangedHandler ElementChangedEvent;
        protected void NotifyElementChanged() {
            if (ElementChangedEvent != null)
                ElementChangedEvent(this);
        }

        public void ExternalNotifyElementChanged() {
            System.Diagnostics.Debug.WriteLine("Element.ExternalNotifyElementChanged called - should figure out how to avoid this!");
            NotifyElementChanged();
        }


        /*
         *  element visibility changed 
         */
        public delegate void ElementVisibilityChangedHandler(Element e);
        public event ElementVisibilityChangedHandler ElementVisibilityChangedEvent;
        protected void NotifyElementVisibilityChanged() {
            if (ElementVisibilityChangedEvent != null)
                ElementVisibilityChangedEvent(this);
        }


        abstract public void RenderBackground();
        abstract public void Render();

#endregion











        /*
         * Utility functions...
         */





        //! determine if item is still renderable
        public static bool VisibleForRendering(VisibilityState state) {
            return (state != VisibilityState.Hidden);
        }

        //! determine if we should incorporate item in layout
        public static bool VisibleForLayout(VisibilityState state) {
            return (state != VisibilityState.Hidden && state != VisibilityState.TransitioningToHidden);
        }

        //! determine visibility of item based on current item and parent state
        public static VisibilityState DetermineVisibility(VisibilityState itemState, VisibilityState parentState) {
            switch (parentState) {
                case VisibilityState.Hidden:
                    return parentState;

                case VisibilityState.TransitioningToHidden:
                    if (itemState == VisibilityState.Hidden)
                        return itemState;
                    else
                        return parentState;

                case VisibilityState.TransitioningToVisible:
                    if (itemState == VisibilityState.Hidden || itemState == VisibilityState.TransitioningToHidden)
                        return itemState;
                    else
                        return parentState;

                case VisibilityState.Visible:
                    if (itemState != VisibilityState.Visible)
                        return itemState;
                    else
                        return parentState;
            }
            return VisibilityState.Hidden;
        }


    }
}
