using System;
using System.Collections.Generic;
using System.Text;

namespace Sketcher
{
    //! A widget is a transient user-interface element (pop-up menu, etc)
    //! Widgets are different from Items because they don't have to be placed
    //! in a Container (in fact, they explicitly can't be), and hence are
    //! not affected by Layout(s)
    abstract class Widget : Element
    {
        public Widget(Element parent) {
            m_transform.PushTransform(Vector2f.ZERO);

            m_fgColor = ColorRGBA.BLACK;
            m_bgColor = ColorRGBA.WHITE;
            m_fOpacity = 1.0f;
            m_Visibility = VisibilityState.Hidden;

            Parent = parent;

            WidgetManager.Get.Register(this);
        }

        ~Widget( ) {
            System.Diagnostics.Debug.WriteLine( "Widget destructor!" );
        }


        //! disconnect widget from the world
        override public void Disconnect( ) {
            base.Disconnect();
            WidgetManager.Get.DeRegister( this );
        }


        //! a Widget should always have a parent Element 
        override public Element Parent {
            set {
                m_parent = value;
                m_transform.Clear();
                if ( m_parent != null )
                    m_transform.PushTransform(m_parent.Transformation);
                m_transform.PushTransform(Vector2f.ZERO);
            }
        }



        //! containment test for widget
        virtual public bool Hit( Vector2f vPoint ) {
            return Bounds.Contains( vPoint );
        }


        //! clipping bounds for widget
        virtual public Box2f ClipRegion {
            get { return Box2f.INFINITE; }
        }


        /*
         * Resizing functions
         */

        virtual public void Resize(Box2f newBounds) {
            Resize(newBounds, (Parent == null) ? null : Parent.Transformation);
        }
        virtual public void Resize(Box2f newBounds, Transform relativeTo) {
            m_size.Width = newBounds.Width;
            m_size.Height = newBounds.Height;
            m_transform.PopTransform();

            // set relative bounds...
            if (relativeTo != null) {
                newBounds = relativeTo.Apply(newBounds, true);
            }

            m_transform.PushTransform(newBounds.Min);
            NotifyElementChanged();
        }


        /*
         * rendering stuff
         */

        //! this can optionally be implemented to draw the widget's background. All
        //! backgrounds will be drawn before foreground. Useful for dropshadows, etc.
        virtual protected void RenderWidgetBackground() {}

        //! this must be implemented to draw widget
        abstract protected void RenderWidget();

        //! widget overrides should update internal geometric elements inside this function (if any)
        virtual protected void UpdateGeometry() {
        }

#region Renderable Members

        override public VisibilityState Visibility {
            get {
                if (Parent != null)
                    return Element.DetermineVisibility(m_Visibility, Parent.Visibility);
                else
                    return m_Visibility;
            }
        }

        override public void RenderBackground() {
            UpdateGeometry();
            RenderWidgetBackground();
        }
        override public void Render() {
            //UpdateGeometry();    (assume that RenderBackground is always called first...)
            RenderWidget();
        }
#endregion

    }
}
