using System;
using System.Collections.Generic;
using System.Text;
using Tao.OpenGl;

namespace Sketcher
{
    class RenderEngine
    {
        // singleton stuff
        static public RenderEngine Get {
            get { 
                if ( s_singleton == null )
                    s_singleton = new RenderEngine();
                return s_singleton; 
                }
        }
        static RenderEngine s_singleton = null;

        public List<Box2f> m_clipBoxStack;
        GlFont2 m_font;
        SplatEngine m_splatEngine;

        Dictionary<Image, GLTexture> m_textures;    //! textures created for images

        protected RenderEngine() {
            m_clipBoxStack = new List<Box2f>();
            m_font = new GlFont2();
            m_textures = new Dictionary<Image,GLTexture>();
            m_splatEngine = new SplatEngine();
        }

        public void Initialize( ) {
            m_splatEngine.Initialize();
        }


        public void DrawText(string s, Vector2f vPos, ColorRGBA color) {
            Gl.glMatrixMode(Gl.GL_PROJECTION);
            Gl.glPushMatrix();
            Gl.glMatrixMode(Gl.GL_MODELVIEW);
            Gl.glPushMatrix();

            int[] nViewport = new int[4];
            Gl.glGetIntegerv(Gl.GL_VIEWPORT, nViewport);
            m_font.Draw((float)nViewport[2], (float)nViewport[3], s, vPos.X, vPos.Y, color);

            Gl.glMatrixMode(Gl.GL_PROJECTION);
            Gl.glPopMatrix();
            Gl.glMatrixMode(Gl.GL_MODELVIEW);
            Gl.glPopMatrix();
        }

        public void DrawBox(Box2f box, ColorRGBA color) {
            Gl.glColor4f(color.R, color.G, color.B, color.A);
            Gl.glBegin(Gl.GL_QUADS);
            Gl.glVertex2f(box.Min.X, box.Min.Y);
            Gl.glVertex2f(box.Max.X, box.Min.Y);
            Gl.glVertex2f(box.Max.X, box.Max.Y);
            Gl.glVertex2f(box.Min.X, box.Max.Y);
            Gl.glEnd();        
        }


        public void DrawBoxOutline( Box2f box, ColorRGBA color ) {
            Gl.glColor4f( color.R, color.G, color.B, color.A );
            Gl.glBegin( Gl.GL_LINE_LOOP );
            Gl.glVertex2f( box.Min.X, box.Min.Y );
            Gl.glVertex2f( box.Max.X, box.Min.Y );
            Gl.glVertex2f( box.Max.X, box.Max.Y );
            Gl.glVertex2f( box.Min.X, box.Max.Y );
            Gl.glEnd();
        }


        Glu.GLUquadric s_quadric;
        bool s_bQuadricAllocated = false;
        Glu.GLUquadric GetQuadric( ) {
            if (!s_bQuadricAllocated) {
                s_quadric = Glu.gluNewQuadric();
                s_bQuadricAllocated = true;
            }
            return s_quadric;
        }


        public void DrawDisc(Vector2f vCenter, float fInnerRadius, float fOuterRadius, ColorRGBA color) {
            Gl.glColor4f(color.R, color.G, color.B, color.A);
            Gl.glPushMatrix();
            Gl.glTranslatef(vCenter.X, vCenter.Y, 0);
            Glu.gluDisk(GetQuadric(), fInnerRadius, fOuterRadius, 40, 1);
            Gl.glPopMatrix();
        }
        public void DrawCircle(Vector2f vCenter, float fRadius, ColorRGBA color) {
            DrawDisc(vCenter, 0, fRadius, color);
        }
        public void DrawCircle(Circle2f circle, ColorRGBA color) {
            DrawCircle(circle.Center, circle.Radius, color);
        }


        public void DrawCircleOutline( Circle2f circle, float fWidth, ColorRGBA color ) {
            DrawDisc( circle.Center, circle.Radius - fWidth, circle.Radius, color );
        }


        // TODO: support color!
        public void DrawLine( Vector2f v1, Vector2f v2, float fWidth, ColorRGBA color ) {
            m_splatEngine.SplatScale = fWidth;
            m_splatEngine.SplatAlpha = color.A;
            m_splatEngine.DrawSplatLine(v1, v2, 1.0f, 1.0f, 1.0f, 1.0f, false);
        }

        public void DrawLine( Vector2f v1, Vector2f v2, float fSize1, float fSize2, float fOpacity1, float fOpacity2, float fWidthScale, ColorRGBA color ) {
            m_splatEngine.SplatScale = fWidthScale;
            m_splatEngine.SplatAlpha = color.A;
            m_splatEngine.DrawSplatLine( v1, v2, fSize1, fSize2, fOpacity1, fOpacity2, false );
        }

        public void DrawStroke(List<Vector2f> vVertices, List<float> vSizes, List<float> vOpacities, float fWidthScale, float fOpacity, ColorRGBA color) {
            m_splatEngine.SplatScale = fWidthScale;
            m_splatEngine.SplatAlpha = fOpacity;
            m_splatEngine.DrawSplatStroke(vVertices, vSizes, vOpacities);
        }




        public void PushClipBox(Box2f box) {
            if (m_clipBoxStack.Count != 0) {
                box = box.Intersect( m_clipBoxStack[m_clipBoxStack.Count-1] );
            }
            m_clipBoxStack.Add(box);

            Gl.glScissor( (int)Math.Floor(box.Min.X), 
                (int)Math.Floor(box.Min.Y), 
                (int)Math.Ceiling(box.Width),
                (int)Math.Ceiling( box.Height ) );
            Gl.glEnable(Gl.GL_SCISSOR_TEST);
        }
        public void PushDisableClipBox( ) {
            m_clipBoxStack.Add( Box2f.INFINITE );
            Gl.glDisable( Gl.GL_SCISSOR_TEST );
        }
        public void PopClipBox() {
            m_clipBoxStack.RemoveAt(m_clipBoxStack.Count-1);
            if (m_clipBoxStack.Count > 0) {
                Box2f box = m_clipBoxStack[m_clipBoxStack.Count - 1];
                Gl.glScissor( (int)Math.Floor( box.Min.X ),
                    (int)Math.Floor( box.Min.Y ),
                    (int)Math.Ceiling( box.Width ),
                    (int)Math.Ceiling( box.Height ) );
                Gl.glEnable(Gl.GL_SCISSOR_TEST);
            } else {
                Gl.glDisable(Gl.GL_SCISSOR_TEST);
            }
        }



        public void PushTransform() {
            Gl.glPushMatrix();
        }
        public void PushClearTransform( ) {
            Gl.glPushMatrix();
            Gl.glLoadIdentity();
        }
        public void PushTransform( Transform t ) {
            Gl.glPushMatrix();
            
            // [TODO] support rotation and scaling
            Gl.glTranslatef( t.Translation.X, t.Translation.Y, 0.0f );
        }
        public void PopTransform( ) {
            Gl.glPopMatrix();
        }
        public void Translate(Vector2f vTranslate) {
            Gl.glTranslatef(vTranslate.X, vTranslate.Y, 0.0f);
        }
        public void Transform( Transform t ) {
            // [TODO] support rotation and scaling
            Gl.glTranslatef( t.Translation.X, t.Translation.Y, 0.0f );
        }


        public void EnableAlphaBlending() {
            Gl.glEnable(Gl.GL_BLEND);
            Gl.glBlendFunc(Gl.GL_SRC_ALPHA, Gl.GL_ONE_MINUS_SRC_ALPHA);
        }

        public void DisableAlphaBlending() {
            Gl.glDisable(Gl.GL_BLEND);
        }


        public void DrawImage( Image image, Box2f box ) {
            if ( ! m_textures.ContainsKey(image) ) {
                GLTexture tex = new GLTexture(image);
                m_textures[image] = tex;
            }
            GLTexture texture = m_textures[image];
            texture.Enable();

            Gl.glBegin(Gl.GL_QUADS);
            Gl.glTexCoord2f(0,0);
            Gl.glVertex2f(box.Min.X, box.Min.Y);
            Gl.glTexCoord2f(1,0);
            Gl.glVertex2f(box.Max.X, box.Min.Y);
            Gl.glTexCoord2f(1,1);
            Gl.glVertex2f(box.Max.X, box.Max.Y);
            Gl.glTexCoord2f(0,1);
            Gl.glVertex2f(box.Min.X, box.Max.Y);
            Gl.glEnd();   

            texture.Disable();
        }



        public Image GrabScreenshot( Box2f bounds ) {
            int[] viewport = {0,0,0,0};
            double[] modelMatrix = new double[16];
            double[] projMatrix = new double[16];
            Gl.glGetIntegerv(Gl.GL_VIEWPORT, viewport);
            Gl.glGetDoublev(Gl.GL_PROJECTION_MATRIX, projMatrix);
            Gl.glGetDoublev(Gl.GL_MODELVIEW_MATRIX, modelMatrix);

            // find pixel bounding box
            Box2f screenBox = new Box2f();
            double[] dScreen = {0,0,0};
            for ( int i = 0; i < 4; ++i ) {
                Vector2f vCorner = bounds.GetCorner(i);
                Glu.gluUnProject( vCorner.X, vCorner.Y, 0, modelMatrix, projMatrix, viewport, 
                    out dScreen[0], out dScreen[1], out dScreen[2] );
                if ( i == 0 )
                    screenBox = new Box2f( 
                        (float)dScreen[0], (float)dScreen[1], (float)dScreen[0], (float)dScreen[1] );
                else
                    screenBox.Contain( new Vector2f((float)dScreen[0], (float)dScreen[1]) );
            }
            int xMin = Math.Max(viewport[0], (int)screenBox.Min.X );
            int yMin = Math.Max(viewport[1], (int)screenBox.Min.Y );
            int xWidth = Math.Min(viewport[2], (int)screenBox.Width );
            int yHeight = Math.Min(viewport[3], (int)screenBox.Height );

            byte[] buffer = new byte[xWidth*yHeight*4];
            Gl.glReadPixels( xMin, yMin, xWidth, yHeight, Gl.GL_RGBA, Gl.GL_UNSIGNED_BYTE, buffer );

            return new Image(xWidth, yHeight, 4, buffer);
        }

    }
}
