using System;
using System.Collections.Generic;
using System.Text;

using Tao.OpenGl;

namespace Sketcher
{
    class SplatEngine
    {
        float m_fSplatScale;
        float m_fSplatAlpha;

        Image m_splatImage;
        GLTexture m_splatTexture;

        bool m_bInFastSplatDraw;

        public SplatEngine() {
            m_fSplatScale = 4.0f;
            m_fSplatAlpha = 1.0f;
            m_bInFastSplatDraw = false;

            m_splatImage = null;
            m_splatTexture = null;
        }
    
        public float SplatScale {
            get { return m_fSplatScale; }
            set { m_fSplatScale = value; }
        }

        public float SplatAlpha {
            get { return m_fSplatAlpha; }
            set { m_fSplatAlpha = value; }
        }


        public void Initialize( ) {
            ValidateTexture();
            if (m_splatTexture != null) {
                m_splatTexture.Enable();
                m_splatTexture.Disable();
            }
        }



        public void DrawSplat( Vector2f v, float fSize, float fOpacity ) {

            fSize *= m_fSplatScale;
            Gl.glColor4f(1, 1, 1, fOpacity * m_fSplatAlpha);

            if (m_bInFastSplatDraw) {

                if (m_splatTexture != null) {
                    Gl.glTexCoord2f(0, 0);
                    Gl.glVertex2f(v.X - fSize, v.Y - fSize);
                    Gl.glTexCoord2f(1, 0);
                    Gl.glVertex2f(v.X + fSize, v.Y - fSize);
                    Gl.glTexCoord2f(1, 1);
                    Gl.glVertex2f(v.X + fSize, v.Y + fSize);
                    Gl.glTexCoord2f(0, 1);
                    Gl.glVertex2f(v.X - fSize, v.Y + fSize);
                } else {
                    Gl.glColor4f( 0, 0, 0, fOpacity * m_fSplatAlpha );
                    Gl.glVertex2f(v.X, v.Y);
                }

            } else {

                ValidateTexture();
                if ( m_splatTexture != null ) {
                    Gl.glPushAttrib( Gl.GL_ENABLE_BIT | Gl.GL_DEPTH_BITS );
                    Gl.glDepthFunc( Gl.GL_ALWAYS );
                    Gl.glEnable( Gl.GL_BLEND );
                    Gl.glBlendFunc( Gl.GL_SRC_ALPHA, Gl.GL_ONE_MINUS_SRC_ALPHA );

                    m_splatTexture.Enable();

                    Gl.glBegin(Gl.GL_QUADS);
                    Gl.glTexCoord2f(0, 0);
                    Gl.glVertex2f(v.X - fSize, v.Y - fSize);
                    Gl.glTexCoord2f(1, 0);
                    Gl.glVertex2f(v.X + fSize, v.Y - fSize);
                    Gl.glTexCoord2f(1, 1);
                    Gl.glVertex2f(v.X + fSize, v.Y + fSize);
                    Gl.glTexCoord2f(0, 1);
                    Gl.glVertex2f(v.X - fSize, v.Y + fSize);
                    Gl.glEnd();

                    m_splatTexture.Disable();

                    Gl.glPopAttrib();

                } else {
                    Gl.glBegin(Gl.GL_POINTS);
                    Gl.glVertex2f(v.X, v.Y);
                    Gl.glEnd();
                }

            }
        }


        bool Begin_Fast_SplatDraw() {

            if (m_bInFastSplatDraw)
                return false;
        
            ValidateTexture();
            if (m_splatTexture != null) {
                Gl.glPushAttrib( Gl.GL_ENABLE_BIT | Gl.GL_DEPTH_BITS );
                Gl.glDepthFunc( Gl.GL_ALWAYS );
                Gl.glEnable( Gl.GL_BLEND );
                Gl.glBlendFunc( Gl.GL_SRC_ALPHA, Gl.GL_ONE_MINUS_SRC_ALPHA );

                m_splatTexture.Enable();

                Gl.glColor4f(1, 1, 1, 1);

                Gl.glBegin(Gl.GL_QUADS);

            } else {
                Gl.glPushAttrib( Gl.GL_ENABLE_BIT | Gl.GL_DEPTH_BITS );
                Gl.glPointSize( m_fSplatScale );
                Gl.glBegin(Gl.GL_POINTS);
            }

            m_bInFastSplatDraw = true;
            return true;
        }


        void End_Fast_SplatDraw() {
            Gl.glEnd();
            if (m_splatTexture != null)
                m_splatTexture.Disable();

            Gl.glPopAttrib();

            m_bInFastSplatDraw = false;
        }




        public void DrawSplatLine(Vector2f v1, Vector2f v2, float fSize1, float fSize2, float fOpacity1, float fOpacity2, bool bSkipLastSplat) {

            bool bEndFastSplatDraw = Begin_Fast_SplatDraw();

            Vector2f d = (v2 - v1);
            float fLength = d.Length;

            float fStepSize = 0.6f * (SplatScale * Math.Max(fSize1, fSize2));
            int nSteps = (int)(fLength / fStepSize) + 1;
            if (nSteps > 1000)
                nSteps = 1000;

            for (int k = 0; k < nSteps; ++k) {
                float fAlpha = (float)k / (float)nSteps;

                Vector2f vPos = (1-fAlpha)*v1 + fAlpha*v2;
                float fSize = (1-fAlpha)*fSize1 + fAlpha*fSize2;
                float fTrans = (1-fAlpha)*fOpacity1 + fAlpha*fOpacity2;

                DrawSplat(vPos, fSize, fTrans);
            }

            if (bEndFastSplatDraw)
                End_Fast_SplatDraw();
        }




        public void DrawSplatStroke(List<Vector2f> vVertices, List<float> vSizes, List<float> vOpacities) {

            int nCount = vVertices.Count;
            if (nCount < 1)
                return;
            if (nCount == 1) {
                DrawSplat(vVertices[0], vSizes[0], vOpacities[0]);
                return;
            }

            bool bEndFastSplatDraw = Begin_Fast_SplatDraw();

            for (int i = 0; i < nCount-1; ++i)
                DrawSplatLine(vVertices[i], vVertices[i+1], vSizes[i], vSizes[i+1], vOpacities[i], vOpacities[i+1], i != nCount-1);

            if (bEndFastSplatDraw)
                End_Fast_SplatDraw();
        }



        void ValidateTexture() {
            if (m_splatTexture != null)
                return;

            string sResourcePath = NotebookAppConstants.TEMP_SplatResourcePath;
            FilePath fp = new FilePath( FilePath.FileLocation.Resource, sResourcePath, "splat.png" );
            m_splatImage = new Image();
            if ( m_splatImage.Load(fp) ) {
                m_splatTexture = new GLTexture( m_splatImage );
                m_splatTexture.TexEnvMode = Gl.GL_MODULATE;
            }
        }
    

    }
}
